import cvxpy as cp
import numpy as np
from loguru import logger


def bilinear(mat, left=None, right=None):
    n, m = mat.shape
    assert n == m, "Matrix must be square"

    if left is None:
        left = np.ones(n)

    if right is None:
        right = np.ones(n)

    return left.T @ (mat @ right)


if __name__ == "__main__":
    n = 3
    A = np.random.randn(n, n)
    Sigma = A @ A.T

    # We minimize the variance of the portfolio
    # Var = min 1/2 * w'Sigma w
    # subject to the constraint that the portfolio is fully invested
    # and that all weights are non-negative.
    # sum w == 1
    # and w >= 0

    # We can _solve this problem using Lagrange multipliers

    # -gamma*(sum w - 1)
    # w - s^2 = 0
    # -theta_1*(w_1 - s_1^2)
    # -theta_2*(w_2 - s_2^2)

    # For now let's drop the constraint of long-only

    # from
    gamma = 1 / bilinear(np.linalg.inv(Sigma), left=np.ones(n), right=np.ones(n))
    w = gamma * np.linalg.solve(Sigma, np.ones(n))
    logger.info(f"Gamma: {gamma}")
    logger.info(f"Weights: {w}")
    logger.info(f"Sum of weights: {np.sum(w)}")
    logger.info(f"Variance: {w.T @ Sigma @ w}")

    # values are in ascending order
    # values, vectors = np.linalg.eigh(Sigma)
    # logger.info(f"Min eigenvalue: {np.min(values)}")

    # vector = vectors[:,0]
    # a = vector / np.sum(vector)
    # logger.info(f"Eigenvalues of Sigma: {values}")
    # logger.info(f"Scaled eigenvector: {a}")
    # print(a.T @ Sigma @ a)
    # print(np.sqrt(values))

    # print((Sigma @ a)/a)
    # MinVar without the non-negativity constraint
    # is equivalent to the inverse of the covariance matrix

    C = np.linalg.cholesky(Sigma)
    xxx = np.linalg.solve(C, np.ones(n))
    a = xxx / np.sum(xxx)

    print(a)
    print(a.T @ Sigma @ a)

    x = cp.Variable(n)

    objective = cp.Minimize(cp.quad_form(x, Sigma))
    parameters = [cp.sum(x) == 1]

    prob = cp.Problem(objective, parameters)
    prob.solve()

    a = x.value
    print(a.T @ Sigma @ a)



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from dataclasses import dataclass

import numpy as np

from ..claux import CLAUX
from ..types import TurningPoint


@dataclass(frozen=True)
class CLA(CLAUX):
    def __post_init__(self):
        ns = self.mean.shape[0]
        m = self.A.shape[0]

        # Initialize the portfolio.
        first = self._first_turning_point()
        self._append(first)

        # Set the P matrix.
        P = np.block([self.covariance, self.A.T])
        M = np.block([[self.covariance, self.A.T], [self.A, np.zeros((m, m))]])

        lam = np.inf

        while lam > 0:
            last = self.turning_points[-1]

            blocked = ~last.free
            assert not np.all(blocked), "Not all variables can be blocked"

            # Create the UP, DN, and IN
            UP = blocked & np.isclose(last.weights, self.upper_bounds)
            DN = blocked & np.isclose(last.weights, self.lower_bounds)

            # a variable is out if it is UP or DN
            OUT = np.logical_or(UP, DN)
            IN = ~OUT

            up = np.zeros(ns)
            up[UP] = self.upper_bounds[UP]

            dn = np.zeros(ns)
            dn[DN] = self.lower_bounds[DN]

            top = np.copy(self.mean)
            top[OUT] = 0

            _IN = np.concatenate([IN, np.ones(m, dtype=np.bool_)])

            bbb = np.array([np.block([up + dn, self.b]), np.block([top, np.zeros(m)])]).T

            alpha, beta = CLA._solve(M, bbb, _IN)

            gamma = P @ alpha
            delta = P @ beta - self.mean

            # Prepare the ratio matrix.
            L = -np.inf * np.ones([ns, 4])

            r_beta = beta[range(ns)]
            r_alpha = alpha[range(ns)]

            # IN security possibly going UP.
            i = IN & (r_beta < -self.tol)
            L[i, 0] = (self.upper_bounds[i] - r_alpha[i]) / r_beta[i]

            # IN security possibly going DN.
            i = IN & (r_beta > +self.tol)
            L[i, 1] = (self.lower_bounds[i] - r_alpha[i]) / r_beta[i]

            # UP security possibly going IN.
            i = UP & (delta < -self.tol)
            L[i, 2] = -gamma[i] / delta[i]

            # DN security possibly going IN.
            i = DN & (delta > +self.tol)
            L[i, 3] = -gamma[i] / delta[i]

            # If all elements of ratio are negative,
            # we have reached the end of the efficient frontier.
            if np.max(L) < 0:
                break

            secchg, dirchg = np.unravel_index(np.argmax(L, axis=None), L.shape)

            # Set the new value of lambda_E.
            lam = L[secchg, dirchg]

            free = np.copy(last.free)
            if dirchg == 0 or dirchg == 1:
                free[secchg] = False
            else:
                free[secchg] = True

            # Compute the portfolio at this corner.
            x = r_alpha + lam * r_beta

            # Save the data computed at this corner.
            self._append(TurningPoint(lamb=lam, weights=x, free=free))

        self._append(TurningPoint(lamb=0, weights=r_alpha, free=last.free))

    @staticmethod
    def _solve(A, b, IN):
        OUT = ~IN
        n = A.shape[1]
        x = np.zeros((n, 2))

        x[OUT, :] = b[OUT, :]
        bbb = b[IN, :] - A[IN, :][:, OUT] @ x[OUT, :]

        x[IN, :] = np.linalg.inv(A[IN, :][:, IN]) @ bbb
        return x[:, 0], x[:, 1]



"""global fixtures"""

from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pandas as pd
import pytest

from cvx.bson import read_bson


@pytest.fixture(scope="session", name="resource_dir")
def resource_fixture():
    """resource fixture"""
    return Path(__file__).parent / "resources"


@pytest.fixture()
def input_data(resource_dir):
    data = read_bson(file=resource_dir / "input_data.bson")
    return SimpleNamespace(**data)


@pytest.fixture()
def results(resource_dir):
    data = read_bson(file=resource_dir / "results.bson")
    return SimpleNamespace(**data)


@pytest.fixture()
def example(resource_dir):
    return pd.read_csv(resource_dir / "example.csv", header=0, index_col=0)


@pytest.fixture()
def example_solution(resource_dir):
    return pd.read_csv(resource_dir / "example_solution.csv", header=0, index_col=None)



import numpy as np

from tests.bailey.cla import CLA


def test_big(input_data, results):
    cla = CLA(
        mean=input_data.mean,
        lower_bounds=input_data.lower_bounds,
        upper_bounds=input_data.upper_bounds,
        covariance=input_data.covariance,
        A=np.ones((1, len(input_data.mean))),
        b=np.ones(1),
    )

    observed = [tp.lamb for tp in cla.turning_points[1:]]
    np.allclose(results.lamb, np.array(observed))

    observed = [tp.mean(input_data.mean) for tp in cla.turning_points[1:]]
    np.allclose(results.mean, np.array(observed))

    observed = [tp.variance(input_data.covariance) for tp in cla.turning_points[1:]]
    np.allclose(results.variance, np.array(observed))

    observed = [tp.weights for tp in cla.turning_points[1:]]
    np.allclose(results.weights, np.array(observed))



import numpy as np
import pytest

from tests.bailey.cla import _Schur


def test_compute_lambda():
    # Create a _Schur instance with some sample data
    covariance = np.array([[1, 0.5, 0.2], [0.5, 2, 0.3], [0.2, 0.3, 3]])
    mean = np.array([1, 2, 3])
    free = np.array([True, True, False])
    weights = np.array([0.3, 0.5, 0.2])
    schur = _Schur(covariance, mean, free, weights)

    # Test the compute_lambda method with different inputs
    index = 1
    bi = np.array([0.5])
    expected_lambda = 0.62
    expected_gamma = 0.5
    actual_lambda, actual_gamma = schur.compute_lambda(index, bi)

    assert np.isclose(actual_lambda, expected_lambda)
    assert np.isclose(actual_gamma, expected_gamma)

    index = 1
    bi = np.array([0.2, 0.3])
    expected_lambda = 0.02
    expected_gamma = 0.2
    actual_lambda, actual_gamma = schur.compute_lambda(index, bi)
    assert np.isclose(actual_lambda, expected_lambda)
    assert np.isclose(actual_gamma, expected_gamma)


def test_compute_weight():
    # Define some test data
    covariance = np.array([[1, 0.5, 0.2], [0.5, 2, 0.3], [0.2, 0.3, 3]])
    mean = np.array([1.0, 1.0, 3.0])
    free = np.array([True, True, False])
    weights = np.array([0.4, 0.6, 0.0])

    # Create a _Schur object
    schur = _Schur(covariance, mean, free, weights)

    # Test the compute_weight method
    lamb = 0.5

    actual_weights = schur.update_weights(lamb)
    np.testing.assert_allclose(actual_weights, np.array([0.75, 0.25, 0.0]), rtol=1e-3)


def test_special_minvar():
    mean = np.array([0.1, 0.2])

    lower_bounds = np.array([0.0, 0.0])
    upper_bounds = np.array([0.6, 0.7])
    covariance = np.array([[2.0, 1.0], [1.0, 3.0]])

    # start with the vector [0.3, 0.7]. Try to free the second asset
    schur = _Schur(
        covariance=covariance,
        mean=mean,
        free=np.array([True, True]),
        weights=np.array([0.3, 0.7]),
    )
    lamb, bi = schur.compute_lambda(index=1, bi=np.array([0.7]))
    assert lamb == pytest.approx(11.0)
    assert bi == 0.7
    w = schur.update_weights(lamb=11)
    assert np.allclose(w, np.array([0.3, 0.7]))

    schur = _Schur(
        covariance=covariance,
        mean=mean,
        free=np.array([True, True]),
        weights=np.array([0.3, 0.7]),
    )

    # try to block the first asset
    lamb, bi = schur.compute_lambda(
        index=0,
        bi=np.array([lower_bounds[0], upper_bounds[0]]),
    )

    assert lamb == pytest.approx(2.0)
    assert bi == pytest.approx(0.6)

    schur = _Schur(
        covariance=covariance,
        mean=mean,
        free=np.array([False, True]),
        weights=np.array([0.6, 0.7]),
    )
    w = schur.update_weights(lamb=2)

    assert np.allclose(w, np.array([0.6, 0.4]))



import numpy as np
from loguru import logger

from cvx.cla.first import init_algo, init_algo_lp

if __name__ == "__main__":
    n = 10000
    mean = np.ones(n)
    upper_bound = np.ones(n)

    logger.info("Hello")

    tp = init_algo(mean=mean, lower_bounds=np.zeros_like(upper_bound), upper_bounds=upper_bound)
    print(np.where(tp.free)[0])

    tp = init_algo_lp(mean=mean, lower_bounds=np.zeros_like(upper_bound), upper_bounds=upper_bound)
    print(np.where(tp.free)[0])



import numpy as np
import pytest

from cvx.cla.types import TurningPoint


@pytest.fixture()
def tp():
    """Fixture for turning point"""
    return TurningPoint(weights=np.array([0.5, 0.5]), free=np.array([True, False]))


def test_turningpoint(tp):
    """Test turning point"""
    assert np.isinf(tp.lamb)
    assert np.allclose(tp.weights, [0.5, 0.5])
    assert np.allclose(tp.free, [True, False])


def test_indices(tp):
    """Test indices. Both free and blocked"""
    assert np.allclose(tp.free_indices, [0])
    assert np.allclose(tp.blocked_indices, [1])


def test_mean(tp):
    """Test mean"""
    x = tp.mean(mean=np.array([1.0, 2.0]))
    assert x == pytest.approx(1.5)


def test_variance(tp):
    """Test variance"""
    x = tp.variance(covariance=np.array([[2.0, 0.2], [0.2, 2.0]]))
    assert x == pytest.approx(1.1)



from mosek.fusion import Domain, Expr, Model, ObjectiveSense

if __name__ == "__main__":
    A = [[3.0, 1.0, 2.0, 0.0], [2.0, 1.0, 3.0, 1.0], [0.0, 2.0, 0.0, 3.0]]
    c = [3.0, 1.0, 5.0, 1.0]

    # Create a model with the name 'lo1'
    with Model("lo1") as M:
        # Create variable 'x' of length 4
        x = M.variable("x", 4, Domain.greaterThan(0.0))

        # Create constraints
        M.constraint(x.index(1), Domain.lessThan(10.0))
        M.constraint("c1", Expr.dot(A[0], x), Domain.equalsTo(30.0))
        M.constraint("c2", Expr.dot(A[1], x), Domain.greaterThan(15.0))
        M.constraint("c3", Expr.dot(A[2], x), Domain.lessThan(25.0))

        # Set the objective function to (c^t * x)
        M.objective("obj", ObjectiveSense.Maximize, Expr.dot(c, x))

        # Solve the problem
        M._solve()

        # Get the solution values
        sol = x.level()
        print("\n".join(["x[%d] = %f" % (i, sol[i]) for i in range(4)]))



import numpy as np
import pytest

from cvx.cla.markowitz.cla import CLA as MARKOWITZ
from tests.bailey.cla import CLA as BAILEY


@pytest.mark.parametrize("solver", [BAILEY, MARKOWITZ])
def test_example(example, example_solution, solver):
    # example from section 3.1 in the Markowitz 2019 paper
    means = example.mean(axis=0)
    std = example.std(axis=0, ddof=1)
    assert np.allclose(means.values, np.array([0.062, 0.146, 0.128]), atol=1e-3)
    assert np.allclose(np.power(std.values, 2), np.array([0.016, 0.091, 0.031]), atol=1e-3)

    ns = example.shape[1]

    lower_bounds = 0.1 * np.ones(ns)
    upper_bounds = 0.5 * np.ones(ns)

    cla = solver(
        mean=means.values,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        covariance=example.cov().values,
        A=np.ones((1, len(means))),
        b=np.ones(1),
    )

    for row, turning_point in enumerate(cla.turning_points):
        if row > 0:
            assert turning_point.lamb == pytest.approx(example_solution["lambda"][row], abs=1e-3)

        assert np.allclose(turning_point.weights, example_solution.values[row, 1:], atol=1e-3)


@pytest.mark.parametrize("solver", [MARKOWITZ, BAILEY])
@pytest.mark.parametrize("n", [2, 4, 8, 16, 32, 64, 128])  # , 256, 512])
def test_init_solver(solver, n):
    mean = np.random.randn(n)
    A = np.random.randn(n, n)
    sigma = 0.1

    solver(
        mean=mean,
        lower_bounds=np.zeros(n),
        upper_bounds=np.ones(n),
        covariance=A @ A.T + sigma * np.eye(n),
        A=np.ones((1, len(mean))),
        b=np.ones(1),
    )



import numpy as np
from loguru import logger


def f(lamb):
    denom = np.sum(np.sum(np.linalg.inv(Sigma), axis=1))
    numerator = 1 - lamb * np.sum(np.linalg.inv(Sigma) @ mu)

    gamma = numerator / denom

    w = lamb * np.linalg.inv(Sigma) @ mu + gamma * np.sum(np.linalg.inv(Sigma), axis=1)
    logger.info("********************************************************")
    logger.info(f"Lambda: {lamb}")
    logger.info(f"Gamma: {gamma}")
    logger.info(f"Weights: {w}")
    logger.info(f"Sum of weights: {np.sum(w)}")
    logger.info(f"Variance: {w.T @ Sigma @ w}")


if __name__ == "__main__":
    # https://www.cs.ubc.ca/~schmidtm/Courses/Notes/linearQuadraticGradients.pdf

    n = 3
    A = np.random.randn(n, n)
    Sigma = A @ A.T

    # ones = np.ones(n)
    mu = np.random.randn(n)

    f(lamb=+0.0)
    f(lamb=+0.1)
    f(lamb=+0.2)
    f(lamb=-0.1)
    f(lamb=-0.2)



import numpy as np
import pytest

from cvx.cla.first import init_algo_lp
from cvx.cla.markowitz.cla import CLA


def test_solver(input_data, results):
    cla = CLA(
        mean=input_data.mean,
        lower_bounds=input_data.lower_bounds,
        upper_bounds=input_data.upper_bounds,
        covariance=input_data.covariance,
        tol=1e-5,
        A=np.array([np.ones_like(input_data.mean)]),
        b=np.ones(1),
    )

    observed = np.array([tp.lamb for tp in cla.turning_points[1:]])
    assert np.allclose(results.lamb, observed, atol=1e-2)

    observed = np.array([tp.mean(input_data.mean) for tp in cla.turning_points[1:]])
    assert np.allclose(results.mean, observed, atol=1e-2)

    observed = np.array([tp.variance(input_data.covariance) for tp in cla.turning_points[1:]])
    assert np.allclose(results.variance, observed, atol=0.5)


def test_example(example, example_solution):
    # example from section 3.1 in the Markowitz 2019 paper
    means = example.mean(axis=0)
    std = example.std(axis=0, ddof=1)
    assert np.allclose(means.values, np.array([0.062, 0.146, 0.128]), atol=1e-3)
    assert np.allclose(np.power(std.values, 2), np.array([0.016, 0.091, 0.031]), atol=1e-3)

    ns = example.shape[1]

    lower_bounds = 0.1 * np.ones(ns)
    upper_bounds = 0.5 * np.ones(ns)

    # covariance = example.cov().values
    tp = init_algo_lp(mean=means.values, lower_bounds=lower_bounds, upper_bounds=upper_bounds)
    assert np.allclose(tp.weights, np.array([0.1, 0.5, 0.4]), atol=1e-9)
    assert np.allclose(tp.free, np.array([False, False, True]))

    cla = CLA(
        mean=means.values,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        covariance=example.cov().values,
        A=np.ones((1, len(means))),
        b=np.ones(1),
    )

    for row, turning_point in enumerate(cla.turning_points):
        if row > 0:
            assert turning_point.lamb == pytest.approx(example_solution["lambda"][row], abs=1e-3)

        assert np.allclose(turning_point.weights, example_solution.values[row, 1:], atol=1e-3)



from dataclasses import dataclass

import numpy as np
import pytest
from loguru import logger

from cvx.cla.claux import CLAUX
from cvx.cla.types import TurningPoint


@dataclass(frozen=True)
class Cla(CLAUX):
    pass


@pytest.fixture()
def cla(input_data):
    return Cla(
        covariance=input_data.covariance,
        mean=input_data.mean,
        lower_bounds=input_data.lower_bounds,
        upper_bounds=input_data.upper_bounds,
        tol=1e-5,
        A=np.ones((1, len(input_data.mean))),
        b=np.ones(1),
        logger=logger,
    )


def test_claux(cla, input_data):
    """
    Test that the CLAUX class is initialized correctly
    """
    np.testing.assert_equal(cla.covariance, input_data.covariance)
    np.testing.assert_equal(cla.mean, input_data.mean)
    np.testing.assert_equal(cla.lower_bounds, input_data.lower_bounds)
    np.testing.assert_equal(cla.upper_bounds, input_data.upper_bounds)
    assert cla.tol == 1e-5


def test_append(cla):
    weights = np.random.rand(10)
    weights = weights / np.sum(weights)

    assert np.all(weights <= cla.upper_bounds)
    assert np.all(weights >= cla.lower_bounds)

    tp = TurningPoint(
        weights=weights,
        free=np.full_like(cla.mean, fill_value=True, dtype=np.bool_),
        lamb=2.0,
    )

    assert tp.lamb == 2.0

    cla._append(tp)

    # assert cla.num_points == 1
    assert len(cla) == 1

    assert cla.turning_points[-1].lamb == 2.0
    # all variables are free
    assert np.all(cla.turning_points[-1].free)

    fig = cla.frontier.plot()
    assert fig is not None


def test_raise():
    cla = Cla(
        covariance=np.eye(2),
        upper_bounds=np.ones(2),
        lower_bounds=np.zeros(2),
        mean=np.ones(2),
        tol=1e-5,
        A=np.ones((1, 2)),
        b=np.ones(1),
    )

    with pytest.raises(AssertionError):
        tp = TurningPoint(
            weights=np.array([0.6, 0.6]),
            free=np.full_like(cla.mean, fill_value=True, dtype=np.bool_),
            lamb=2.0,
        )
        cla._append(tp)

    with pytest.raises(AssertionError):
        tp = TurningPoint(
            weights=np.array([1.2, 0.6]),
            free=np.full_like(cla.mean, fill_value=True, dtype=np.bool_),
            lamb=2.0,
        )
        cla._append(tp)

    with pytest.raises(AssertionError):
        tp = TurningPoint(
            weights=np.array([0.6, -0.6]),
            free=np.full_like(cla.mean, fill_value=True, dtype=np.bool_),
            lamb=2.0,
        )
        cla._append(tp)



##
# Copyright: Copyright (c) MOSEK ApS, Denmark. All rights reserved.
#
# File:      portfolio_2_frontier.py
#
#  Purpose :   Implements a basic portfolio optimization model.
#              Computes points on the efficient frontier.
#
##


import numpy as np
from mosek.fusion import Domain, Expr, Model, ObjectiveSense

"""
Purpose:
    Computes several portfolios on the optimal portfolios by

        for alpha in alphas:
            maximize   expected return - alpha * variance
            subject to the constraints

Input:
    n: Number of assets
    mu: An n dimensional vector of expected returns
    GT: A matrix with n columns so (GT')*GT  = covariance matrix
    x0: Initial holdings
    w: Initial cash holding
    alphas: List of the alphas

Output:
    The efficient frontier as list of tuples (alpha, expected return, variance)
"""


def EfficientFrontier(n, mu, GT, x0, w, alphas):
    with Model("Efficient frontier") as M:
        frontier = []

        # Defines the variables (holdings). Shortselling is not allowed.
        x = M.variable("x", n, Domain.greaterThan(0.0))  # Portfolio variables
        s = M.variable("s", 1, Domain.unbounded())  # Variance variable

        # Total budget constraint
        M.constraint("budget", Expr.sum(x), Domain.equalsTo(w + sum(x0)))

        # Computes the risk
        M.constraint("variance", Expr.vstack(s, 0.5, Expr.mul(GT, x)), Domain.inRotatedQCone())

        # Define objective as a weighted combination of return and variance
        alpha = M.parameter()
        M.objective(
            "obj",
            ObjectiveSense.Maximize,
            Expr.sub(Expr.dot(mu, x), Expr.mul(alpha, s)),
        )

        # Solve multiple instances by varying the parameter alpha
        for a in alphas:
            alpha.setValue(a)

            M._solve()

            frontier.append((a, np.dot(mu, x.level()), s.level()[0]))

        return frontier


if __name__ == "__main__":
    n = 8
    w = 1.0
    mu = [0.07197, 0.15518, 0.17535, 0.08981, 0.42896, 0.39292, 0.32171, 0.18379]
    x0 = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    GT = [
        [0.30758, 0.12146, 0.11341, 0.11327, 0.17625, 0.11973, 0.10435, 0.10638],
        [0.0, 0.25042, 0.09946, 0.09164, 0.06692, 0.08706, 0.09173, 0.08506],
        [0.0, 0.0, 0.19914, 0.05867, 0.06453, 0.07367, 0.06468, 0.01914],
        [0.0, 0.0, 0.0, 0.20876, 0.04933, 0.03651, 0.09381, 0.07742],
        [0.0, 0.0, 0.0, 0.0, 0.36096, 0.12574, 0.10157, 0.0571],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.21552, 0.05663, 0.06187],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.22514, 0.03327],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2202],
    ]

    # Some predefined alphas are chosen
    alphas = [
        0.0,
        0.01,
        0.1,
        0.25,
        0.30,
        0.35,
        0.4,
        0.45,
        0.5,
        0.75,
        1.0,
        1.5,
        2.0,
        3.0,
        10.0,
    ]
    frontier = EfficientFrontier(n, mu, GT, x0, w, alphas)
    print("\n-----------------------------------------------------------------------------------")
    print("Efficient frontier")
    print("-----------------------------------------------------------------------------------\n")
    print("%-12s  %-12s  %-12s" % ("alpha", "return", "risk (std. dev.)"))
    for i in frontier:
        print("{:<12.4f}  {:<12.4e}  {:<12.4e}".format(i[0], i[1], np.sqrt(i[2])))



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from dataclasses import dataclass

import numpy as np
from loguru import logger

from cvx.cla.claux import CLAUX
from cvx.cla.types import BOOLEAN_VECTOR, MATRIX, TurningPoint


@dataclass(frozen=True)
class CLA(CLAUX):
    def __post_init__(self):
        # Compute the turning points,free sets and weights
        self._append(self._first_turning_point())

        while True:
            last = self.turning_points[-1]

            # 1) case a): Bound one free weight
            l_in = -np.inf

            # only try to bound a free asset if there are least two of them
            if np.sum(last.free) > 1:
                schur = _Schur(
                    covariance=self.covariance,
                    mean=self.mean,
                    free=last.free,
                    weights=last.weights,
                )

                for i in last.free_indices:
                    # count the number of entries that are True below the ith entry in fff
                    j = np.sum(last.free[:i])

                    lamb, bi = schur.compute_lambda(
                        index=j,
                        bi=np.array([self.lower_bounds[i], self.upper_bounds[i]]),
                    )

                    if lamb > l_in:
                        l_in, i_in, bi_in = lamb, i, bi

            # 2) case b): Free one bounded weight
            l_out = -np.inf

            for i in last.blocked_indices:
                fff = np.copy(last.free)
                fff[i] = True

                schur = _Schur(
                    covariance=self.covariance,
                    mean=self.mean,
                    free=fff,
                    weights=last.weights,
                )

                # count the number of entries that are True below the ith entry in fff
                j = np.sum(fff[:i])

                lamb, bi = schur.compute_lambda(
                    index=j,
                    bi=np.array([last.weights[i]]),
                )

                if last.lamb > lamb > l_out:
                    l_out, i_out = lamb, i

            l_current = np.max([l_in, l_out])

            if l_current > 0:
                # 4) decide lambda
                logger.info(f"l_in: {l_in}")
                logger.info(f"l_out: {l_out}")
                logger.info(f"l_current: {l_current}")
                f = np.copy(last.free)
                w = np.copy(last.weights)

                if l_in > l_out:
                    lll = l_in
                    f[i_in] = False
                    w[i_in] = bi_in  # set value at the correct boundary
                else:
                    lll = l_out
                    f[i_out] = True
            else:
                break

            schur = _Schur(
                covariance=self.covariance,
                mean=self.mean,
                free=f,
                weights=w,
            )
            # 5) compute solution vector
            weights = schur.update_weights(lamb=lll)
            tp = TurningPoint(weights=weights, lamb=lll, free=f)

            # check the turning point
            self._append(tp)

            logger.info(f"weights: {tp.weights}")
            logger.info(f"free: {tp.free_indices}")

        # 6) compute minimum variance solution
        last = self.turning_points[-1]

        schur = _Schur(
            covariance=self.covariance,
            mean=self.mean,
            free=last.free,
            weights=last.weights,
        )
        w = schur.update_weights(lamb=0)
        # assert np.allclose(x, w, atol=1e-4)

        self._append(TurningPoint(lamb=0, weights=w, free=last.free))


class _Schur:
    def __init__(self, covariance, mean, free: BOOLEAN_VECTOR, weights: MATRIX):
        assert covariance.shape[0] == covariance.shape[1] == mean.shape[0] == free.shape[0] == weights.shape[0]
        self.covariance = covariance
        self.mean = mean
        self.free = free
        self.weights = weights
        self.__free_inv = np.linalg.inv(self.covariance_free)

    @property
    def covariance_free(self):
        return self.covariance[self.free][:, self.free]

    @property
    def covariance_free_blocked(self):
        return self.covariance[self.free][:, ~self.free]

    @property
    def covariance_free_inv(self):
        return self.__free_inv

    @property
    def mean_free(self):
        return self.mean[self.free]

    @property
    def weights_blocked(self):
        return self.weights[~self.free]

    def compute_lambda(self, index, bi):
        def compute_bi(c, bi):
            if np.shape(bi)[0] == 1 or c <= 0:
                return bi[0]
            return bi[1]

        c4 = np.sum(self.covariance_free_inv, axis=0)
        c1 = np.sum(c4)
        c2 = self.covariance_free_inv @ self.mean_free
        # c3 = np.sum(c2)
        # c3 = np.sum(self.covariance_free_inv, axis=1) @ self.mean_free
        # c4 = np.sum(self.covariance_free_inv, axis=0)

        aux = -np.sum(c4) * c2[index] + np.sum(c2) * c4[index]

        bi = compute_bi(aux, bi)

        if self.weights_blocked.size == 0:
            return float((c4[index] - c1 * bi) / aux), bi

        l1 = np.sum(self.weights_blocked)
        l2 = self.covariance_free_inv @ self.covariance_free_blocked
        l3 = l2 @ self.weights_blocked
        l2 = np.sum(l3)
        return ((1 - l1 + l2) * c4[index] - c1 * (bi + l3[index])) / aux, bi

    def _compute_weight(self, lamb):
        g1 = np.sum(self.covariance_free_inv @ self.mean_free, axis=0)
        g2 = np.sum(np.sum(self.covariance_free_inv))

        if self.weights_blocked.size == 0:
            gamma = -lamb * g1 / g2 + 1 / g2
            w1 = 0
        else:
            g3 = np.sum(self.weights_blocked)
            g4 = self.covariance_free_inv @ self.covariance_free_blocked
            w1 = g4 @ self.weights_blocked
            g4 = np.sum(w1)
            gamma = -lamb * g1 / g2 + (1 - g3 + g4) / g2

        w2 = np.sum(self.covariance_free_inv, axis=1)
        w3 = self.covariance_free_inv @ self.mean_free
        return -w1 + gamma * w2 + lamb * w3, gamma

    def update_weights(self, lamb):
        weights, _ = self._compute_weight(lamb)
        new_weights = np.copy(self.weights)
        new_weights[self.free] = weights
        return new_weights



import numpy as np
import pytest

from cvx.cla.first import init_algo, init_algo_lp
from cvx.cla.types import TurningPoint


@pytest.mark.parametrize("n", [5, 20, 100, 1000, 10000])
def test_init_algo(n):
    """Compute a first turning point"""
    mean = np.random.randn(n)
    lower_bounds = np.zeros(n)
    upper_bounds = np.random.rand(n)

    first = init_algo_lp(mean=mean, lower_bounds=lower_bounds, upper_bounds=upper_bounds)

    assert np.sum(first.free) == 1
    assert np.sum(first.weights) == pytest.approx(1.0)
    assert isinstance(first, TurningPoint)

    first = init_algo(mean=mean, lower_bounds=lower_bounds, upper_bounds=upper_bounds)

    assert np.sum(first.free) == 1
    assert np.sum(first.weights) == pytest.approx(1.0)
    assert isinstance(first, TurningPoint)


def test_small():
    """
    Test a first turning point
    """
    mean = np.array([1.0, 2.0, 3.0])
    lower_bound = np.array([0.0, 0.0, 0.0])
    upper_bound = np.array([0.4, 0.4, 0.4])
    tp = init_algo_lp(mean=mean, lower_bounds=lower_bound, upper_bounds=upper_bound)

    assert np.allclose(tp.weights, [0.2, 0.4, 0.4])
    assert tp.lamb == np.inf
    assert np.allclose(tp.free, [True, False, False])

    tp = init_algo(mean=mean, lower_bounds=lower_bound, upper_bounds=upper_bound)

    assert np.allclose(tp.weights, [0.2, 0.4, 0.4])
    assert tp.lamb == np.inf
    assert np.allclose(tp.free, [True, False, False])


def test_no_fully_invested_portfolio():
    """
    Test that the algorithm fails if no fully invested portfolio
    can be constructed
    """
    mean = np.array([1.0, 2.0, 3.0])
    lower_bounds = np.array([0.0, 0.0, 0.0])
    upper_bounds = np.array([0.2, 0.2, 0.2])

    with pytest.raises(ValueError, match="Could not construct a fully invested portfolio"):
        init_algo_lp(mean=mean, lower_bounds=lower_bounds, upper_bounds=upper_bounds)

    with pytest.raises(ValueError, match="Could not construct a fully invested portfolio"):
        init_algo(mean=mean, lower_bounds=lower_bounds, upper_bounds=upper_bounds)


def test_lb_ub_mixed():
    """
    Test that the algorithm fails if lower bounds are greater than upper bounds
    """
    upper_bounds = np.zeros(3)
    lower_bounds = np.ones(3)
    mean = np.ones(3)

    with pytest.raises(ValueError):
        init_algo_lp(mean=mean, lower_bounds=lower_bounds, upper_bounds=upper_bounds)

    with pytest.raises(ValueError):
        init_algo(mean=mean, lower_bounds=lower_bounds, upper_bounds=upper_bounds)



import numpy as np

np.random.seed(42)


def test_last_diagonal():
    """test last diagonal element"""
    a = np.zeros((3, 3))
    a[2, 2] = 1.0

    a = np.zeros((3, 3))
    a[-1, -1] = 2.0

    assert a[-1, -1] == 2.0


def test_last_column():
    """test last column of a numpy array"""
    a = np.zeros((3, 3))
    a[:, 2] = 1.0

    a = np.zeros((3, 3))
    a[:, -1] = 2.0

    a[:-1, -1] = 3.0

    np.testing.assert_array_equal(a[:, -1], np.array([3.0, 3.0, 2.0]))



from mosek.fusion import Domain, Expr, Model, ObjectiveSense

from cvx.bson import read_bson

if __name__ == "__main__":
    data = read_bson(file="data/input_data.bson")

    # Create a model with the name 'lo1'
    with Model("max_expected_return") as M:
        # Create variable 'x' of length 4
        x = M.variable(
            "x",
            len(data["mean"]),
            Domain.inRange(data["lower_bounds"], data["upper_bounds"]),
        )

        M.constraint("fully-invested", Expr.sum(x), Domain.equalsTo(1.0))

        M.objective("expected_return", ObjectiveSense.Maximize, Expr.dot(data["mean"], x))

        # Solve the problem
        M._solve()

        # Get the solution values
        sol = x.level()

        print(sol)



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from __future__ import annotations

import cvxpy as cp
import numpy as np

from .types import MATRIX, TurningPoint


#
def init_algo(mean: MATRIX, lower_bounds: MATRIX, upper_bounds: MATRIX) -> TurningPoint:
    """The key insight behind Markowitz’s CLA is to find first the
    turning point associated with the highest expected return, and then
    compute the sequence of turning points, each with a lower expected
    return than the previous. That first turning point consists in the
    smallest subset of assets with highest return such that the sum of
    their upper boundaries equals or exceeds one.

    We sort the expected returns in descending order.
    This gives us a sequence for searching for the
    first free asset. All weights are initially set to their lower bounds,
    and following the sequence from the previous step, we move those
    weights from the lower to the upper bound until the sum of weights
    exceeds one. If possible the last iterated weight is then reduced
    to comply with the constraint that the sum of weights equals one.
    This last weight is the first free asset,
    and the resulting vector of weights the first turning point.
    """

    if np.any(lower_bounds > upper_bounds):
        raise ValueError("Lower bounds must be less than or equal to upper bounds")

    # Initialize weights to lower bounds
    weights = np.copy(lower_bounds)
    free = np.full_like(mean, False, dtype=np.bool_)

    # Move weights from lower to upper bound
    # until sum of weights hits or exceeds 1
    for index in np.argsort(-mean):
        weights[index] += np.min([upper_bounds[index] - lower_bounds[index], 1.0 - np.sum(weights)])
        if np.sum(weights) >= 1:
            free[index] = True
            break

    # free = _free(weights, lower_bounds, upper_bounds)

    if not np.any(free):
        #    # We have not reached the sum of weights of 1...
        raise ValueError("Could not construct a fully invested portfolio")

    # Return first turning point, the point with the highest expected return.
    return TurningPoint(free=free, weights=weights)


def init_algo_lp(
    mean: MATRIX,
    lower_bounds: MATRIX,
    upper_bounds: MATRIX,
    A_eq: MATRIX | None = None,
    b_eq: MATRIX | None = None,
    solver=cp.CLARABEL,
    **kwargs,
    # A_ub: MATRIX | None = None,
    # b_ub: MATRIX | None = None,
) -> TurningPoint:
    if A_eq is None:
        A_eq = np.atleast_2d(np.ones_like(mean))

    if b_eq is None:
        b_eq = np.array([1.0])

    # if A_ub is None:
    #    A_ub = np.atleast_2d(np.zeros_like(mean))

    # if b_ub is None:
    #    b_ub = np.array([0.0])

    w = cp.Variable(mean.shape[0], "weights")

    objective = cp.Maximize(mean.T @ w)
    constraints = [
        A_eq @ w == b_eq,
        # A_ub @ w <= b_ub,
        lower_bounds <= w,
        w <= upper_bounds,
        cp.sum(w) == 1.0,
    ]

    problem = cp.Problem(objective, constraints)
    problem.solve(solver=solver, **kwargs)
    # check status of problem is optimal
    if problem.status != cp.OPTIMAL:
        raise ValueError("Could not construct a fully invested portfolio")

    # assert problem.status == cp.OPTIMAL
    # print(problem.status)
    # print(status)

    w = w.value

    # compute the distance from the closest bound
    # distance = np.min(
    #    np.array([np.abs(w - lower_bounds), np.abs(upper_bounds - w)]), axis=0
    # )

    # which element has the largest distance to any bound?
    # Even if all assets are at their bounds,
    # we get a (somewhat random) free asset.
    # index = np.argmax(distance)

    # free = np.full_like(mean, False, dtype=np.bool_)
    # free[index] = True

    free = _free(w, lower_bounds, upper_bounds)

    return TurningPoint(free=free, weights=w)


def _free(w, lower_bounds, upper_bounds):
    distance = np.min(np.array([np.abs(w - lower_bounds), np.abs(upper_bounds - w)]), axis=0)

    index = np.argmax(distance)

    free = np.full_like(w, False, dtype=np.bool_)
    free[index] = True
    return free



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""
Critical Line Algorithm auxiliary class implementation.

This module provides the CLAUX class which implements the core functionality
for the Critical Line Algorithm (CLA) used to compute the efficient frontier
in portfolio optimization problems.
"""

import logging
from dataclasses import dataclass, field
from typing import List, Optional

import numpy as np

from .first import init_algo
from .types import MATRIX, Frontier, FrontierPoint, TurningPoint


@dataclass(frozen=True)
class CLAUX:
    """
    Critical Line Algorithm auxiliary class.

    This class implements the core functionality for the Critical Line Algorithm (CLA)
    used to compute the efficient frontier in portfolio optimization problems.

    Attributes:
        mean: Vector of expected returns for each asset.
        covariance: Covariance matrix of asset returns.
        lower_bounds: Vector of lower bounds for asset weights.
        upper_bounds: Vector of upper bounds for asset weights.
        A: Matrix for linear equality constraints (Ax = b).
        b: Vector for linear equality constraints (Ax = b).
        turning_points: List of turning points on the efficient frontier.
        tol: Tolerance for numerical calculations.
        logger: Logger instance for logging information and errors.
    """

    mean: MATRIX
    covariance: MATRIX
    lower_bounds: MATRIX
    upper_bounds: MATRIX
    A: MATRIX
    b: MATRIX
    turning_points: List[TurningPoint] = field(default_factory=list)
    tol: float = 1e-5
    logger: logging.Logger = logging.getLogger(__name__)

    def __len__(self) -> int:
        """
        Get the number of turning points in the efficient frontier.

        Returns:
            The number of turning points currently stored in the object.
        """
        return len(self.turning_points)

    def _first_turning_point(self) -> TurningPoint:
        """
        Calculate the first turning point on the efficient frontier.

        This method uses the init_algo function to find the first turning point
        based on the mean returns and the bounds on asset weights.

        Returns:
            A TurningPoint object representing the first point on the efficient frontier.
        """
        first = init_algo(
            mean=self.mean,
            lower_bounds=self.lower_bounds,
            upper_bounds=self.upper_bounds,
        )
        return first

    def _append(self, tp: TurningPoint, tol: Optional[float] = None) -> None:
        """
        Append a turning point to the list of turning points.

        This method validates that the turning point satisfies the constraints
        before adding it to the list.

        Args:
            tp: The turning point to append.
            tol: Tolerance for constraint validation. If None, uses the class's tol attribute.

        Raises:
            AssertionError: If the turning point violates any constraints.
        """
        tol = tol or self.tol

        assert np.all(tp.weights >= (self.lower_bounds - tol)), f"{(tp.weights + tol) - self.lower_bounds}"
        assert np.all(tp.weights <= (self.upper_bounds + tol)), f"{(self.upper_bounds + tol) - tp.weights}"
        assert np.allclose(np.sum(tp.weights), 1.0), f"{np.sum(tp.weights)}"

        self.turning_points.append(tp)

    @property
    def frontier(self) -> Frontier:
        """
        Get the efficient frontier constructed from the turning points.

        This property creates a Frontier object from the list of turning points,
        which can be used to analyze the risk-return characteristics of the
        efficient portfolios.

        Returns:
            A Frontier object representing the efficient frontier.
        """
        return Frontier(
            covariance=self.covariance,
            mean=self.mean,
            frontier=[FrontierPoint(point.weights) for point in self.turning_points],
        )



#    Copyright 2023 Stanford University Convex Optimization Group
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
"""
types
"""

from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np
import plotly.express as px
from numpy.typing import NDArray
from scipy.optimize import minimize
from typing_extensions import TypeAlias

MATRIX: TypeAlias = NDArray[np.float64]
BOOLEAN_VECTOR: TypeAlias = NDArray[np.bool_]


@dataclass(frozen=True)
class FrontierPoint:
    weights: MATRIX

    def __post_init__(self):
        # check that the sum is close to 1
        assert np.isclose(np.sum(self.weights), 1.0)

    def mean(self, mean: MATRIX):
        """
        Computes the expected return for a turning point
        """
        return float(mean.T @ self.weights)

    def variance(self, covariance: MATRIX):
        """
        Computes the expected variance for a turning point
        """
        return float(self.weights.T @ covariance @ self.weights)


@dataclass(frozen=True)
class TurningPoint(FrontierPoint):
    """
    A turning point is a vector of weights, a lambda value, and a boolean vector
    indicating which assets are free. All assets that are not free are blocked.
    """

    free: BOOLEAN_VECTOR
    lamb: float = np.inf

    @property
    def free_indices(self):
        """
        Returns the indices of the free assets
        """
        return np.where(self.free)[0]

    @property
    def blocked_indices(self):
        """
        Returns the indices of the blocked assets
        """
        return np.where(~self.free)[0]


@dataclass(frozen=True)
class Frontier:
    """
    A frontier is a list of frontier points. Some of them might be turning points.
    """

    mean: MATRIX
    covariance: MATRIX
    frontier: list[FrontierPoint] = field(default_factory=list)

    def interpolate(self, num=100):
        """
        Interpolate the frontier with num-1 points between adjacent frontier points
        Args:
            num: The number of new points in between each pair

        Returns:
            A frontier with many new points
        """

        def _interpolate():
            for w_right, w_left in zip(self.weights[0:-1], self.weights[1:]):
                for lamb in np.linspace(0, 1, num):
                    if lamb > 0:
                        yield FrontierPoint(weights=lamb * w_left + (1 - lamb) * w_right)

        points = list(_interpolate())
        return Frontier(frontier=points, mean=self.mean, covariance=self.covariance)

    def __iter__(self):
        """
        Iterator for all frontier points
        """
        yield from self.frontier

    def __len__(self):
        """
        Number of frontier points
        """
        return len(self.frontier)

    @property
    def weights(self):
        """
        Matrix of weights. One row per point
        """
        return np.array([point.weights for point in self])

    @property
    def returns(self):
        """
        Vector of expected returns.
        """
        return np.array([point.mean(self.mean) for point in self])

    @property
    def variance(self):
        """
        Vector of expected variances.
        """
        return np.array([point.variance(self.covariance) for point in self])

    @property
    def sharpe_ratio(self):
        """
        Vector of expected Sharpe ratios.
        """
        return self.returns / self.volatility

    @property
    def volatility(self):
        """
        Vector of expected volatilities.
        """
        return np.sqrt(self.variance)

    @property
    def max_sharpe(self):
        """
        Maximal Sharpe ratio on the frontier

        Returns:
            Tuple of maximal Sharpe ratio and the weights to achieve it
        """

        def neg_sharpe(alpha, w_left, w_right):
            # convex combination of left and right weights
            weight = alpha * w_left + (1 - alpha) * w_right
            # compute the variance
            var = weight.T @ self.covariance @ weight
            returns = self.mean.T @ weight
            return -returns / np.sqrt(var)

        sharpe_ratios = self.sharpe_ratio

        # in which point is the maximal Sharpe ratio?
        sr_position_max = np.argmax(self.sharpe_ratio)

        # np.min only there for security...
        right = np.min([sr_position_max + 1, len(self) - 1])
        left = np.max([0, sr_position_max - 1])

        # Look to the left and look to the right

        if right > sr_position_max:
            out = minimize(
                neg_sharpe,
                0.5,
                args=(self.weights[sr_position_max], self.weights[right]),
                bounds=((0, 1),),
            )
            var = out["x"][0]
            w_right = var * self.weights[sr_position_max] + (1 - var) * self.weights[right]
            sharpe_ratio_right = -out["fun"]
        else:
            w_right = self.weights[sr_position_max]
            sharpe_ratio_right = sharpe_ratios[sr_position_max]

        if left < sr_position_max:
            out = minimize(
                neg_sharpe,
                0.5,
                args=(self.weights[left], self.weights[sr_position_max]),
                bounds=((0, 1),),
            )
            var = out["x"][0]
            w_left = var * self.weights[left] + (1 - var) * self.weights[sr_position_max]
            sharpe_ratio_left = -out["fun"]
        else:
            w_left = self.weights[sr_position_max]
            sharpe_ratio_left = sharpe_ratios[sr_position_max]

        if sharpe_ratio_left > sharpe_ratio_right:
            return sharpe_ratio_left, w_left

        return sharpe_ratio_right, w_right

    def plot(self, volatility=False, markers=True):
        """
        Plot the frontier
        """
        if not volatility:
            fig = px.line(
                x=self.variance,
                y=self.returns,
                markers=markers,
                labels={"x": "Expected variance", "y": "Expected Return"},
            )
        else:
            fig = px.line(
                x=self.volatility,
                y=self.returns,
                markers=markers,
                labels={"x": "Expected volatility", "y": "Expected Return"},
            )
        return fig



from __future__ import annotations

import numpy as np
import pytest

from cvx.cla.markowitz.cla import CLA as MARKOWITZ
from tests.bailey.cla import CLA as BAILEY

# np.random.seed(40)


@pytest.mark.parametrize("solver", [BAILEY, MARKOWITZ])
def test_frontier(input_data, solver):
    """
    Test the frontier both for Bailey and Markowitz
    """
    f = solver(
        covariance=input_data.covariance,
        mean=input_data.mean,
        lower_bounds=input_data.lower_bounds,
        upper_bounds=input_data.upper_bounds,
        A=np.ones((1, len(input_data.mean))),
        b=np.ones(1),
    ).frontier

    np.testing.assert_equal(f.covariance, input_data.covariance)
    np.testing.assert_almost_equal(f.max_sharpe[0], 4.4535334766464025, decimal=5)
    np.testing.assert_almost_equal(f.mean, input_data.mean)

    g = f.interpolate(num=10)
    np.testing.assert_equal(g.covariance, input_data.covariance)
    np.testing.assert_almost_equal(g.max_sharpe[0], 4.4535334766464025, decimal=5)
    np.testing.assert_almost_equal(g.mean, input_data.mean)


@pytest.mark.parametrize("n", [3, 5, 10, 20])
def test_frontiers(n, resource_dir):
    """
    Compare the frontiers of BAILEY and MARKOWITZ for a variety of dimensions.
    """
    mean = np.random.randn(n)
    lower_bounds = np.zeros(n)
    upper_bounds = np.ones(n)

    cov = np.random.randn(n, n)

    covar = cov @ cov.T

    f_bailey = BAILEY(
        covariance=covar,
        mean=mean,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        A=np.ones((1, n)),
        b=np.ones(1),
        tol=1e-5,
    ).frontier

    f_markowitz = MARKOWITZ(
        mean=np.copy(mean),
        lower_bounds=np.copy(lower_bounds),
        upper_bounds=np.copy(upper_bounds),
        covariance=np.copy(covar),
        tol=1e-5,
        A=np.ones((1, n)),
        b=np.ones(1),
    ).frontier

    assert np.sum(f_bailey.frontier[-1].weights) == pytest.approx(1)
    assert np.sum(f_markowitz.frontier[-1].weights) == pytest.approx(1)
    assert f_bailey.max_sharpe[0] == pytest.approx(f_markowitz.max_sharpe[0])



import marimo

__generated_with = "0.9.27"
app = marimo.App()


@app.cell
def __(mo):
    mo.md(r"""# Critical Line Algorithm""")
    return


@app.cell
def __(mo):
    mo.md(r"""We compute an efficient frontier using the critical line algorithm (cla)""")
    return


@app.cell
def __():
    import numpy as np

    from cvx.cla.markowitz.cla import CLA as MARKOWITZ

    return MARKOWITZ, np


@app.cell
def __(MARKOWITZ, np):
    n = 10
    mean = np.random.randn(n)
    lower_bounds = np.zeros(n)
    upper_bounds = np.ones(n)

    factor = np.random.randn(n, n)
    covariance = factor @ factor.T

    f1 = MARKOWITZ(
        mean=mean,
        covariance=covariance,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        A=np.ones((1, len(mean))),
        b=np.ones(1),
    ).frontier
    return covariance, f1, factor, lower_bounds, mean, n, upper_bounds


@app.cell
def __(f1):
    f1.interpolate(10).plot(volatility=True, markers=False)
    return


@app.cell
def __(f1):
    f1.plot()
    return


@app.cell
def __():
    import marimo as mo

    return (mo,)


if __name__ == "__main__":
    app.run()
